#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.util import MaybeRelative, FrozenOrderedSet


class TestMaybeRelative(unittest.TestCase):
    def test_init_from_float(self):
        arg = MaybeRelative(24.5)
        self.assertFalse(arg.relative)
        self.assertIsNone(arg.reference)
        self.assertAlmostEqual(float(arg), 24.5)

    def test_init_from_float_string(self):
        arg = MaybeRelative('-10')
        self.assertFalse(arg.relative)
        self.assertIsNone(arg.reference)
        self.assertAlmostEqual(float(arg), -10.0)

    def test_init_from_percentage(self):
        arg = MaybeRelative('110%')
        self.assertTrue(arg.relative)
        self.assertIsNone(arg.reference)
        with self.assertRaises(ValueError):
            float(arg)

    def test_resolve_relative(self):
        arg = MaybeRelative('40%')
        arg.reference = 200.0
        self.assertAlmostEqual(float(arg), 80.0)

    def test_invalid(self):
        with self.assertRaises(ValueError):
            arg = MaybeRelative('abc')


class TestFrozenOrderedSet(unittest.TestCase):
    def test_init_from_empty(self):
        s = FrozenOrderedSet()
        self.assertEqual(list(s), [])

    def test_init_from_sequence(self):
        s = FrozenOrderedSet([5, 2, 4, 0, 9, 13, 5, 3])
        self.assertEqual(list(s), [5, 2, 4, 0, 9, 13, 3])

    def test_set_length(self):
        s = FrozenOrderedSet([1, 2, 3])
        self.assertEqual(len(s), 3)

    def test_set_is_hashable(self):
        s = FrozenOrderedSet(['a', 'b', 'c'])
        hash(s)


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

from __future__ import print_function

import re

from ..command import Command, FilePrefixAppendAction
from ..reaction import Compound


def parse_compound(s):
    """Parse a compound specification with optional compartment"""
    m = re.match(r'([^\[]+)\[(\w+)\]', s)
    if m is not None:
        return Compound(m.group(1), compartment=m.group(2))
    return Compound(s)


def filter_search_term(s):
    return re.sub(r'[^a-z0-9]+', '', s.lower())


class SearchCommand(Command):
    """Search for reactions and compounds in the model."""

    @classmethod
    def init_parser(cls, parser):
        """Initialize argument parser"""
        subparsers = parser.add_subparsers(title='Search domain')

        # Compound subcommand
        parser_compound = subparsers.add_parser(
            'compound', help='Search in compounds')
        parser_compound.set_defaults(which='compound')
        parser_compound.add_argument(
            '--id', '-i', dest='id', metavar='id',
            action=FilePrefixAppendAction, type=str, default=[],
            help='Compound ID')
        parser_compound.add_argument(
            '--name', '-n', dest='name', metavar='name',
            action=FilePrefixAppendAction, type=str, default=[],
            help='Name of compound')

        # Reaction subcommand
        parser_reaction = subparsers.add_parser(
            'reaction', help='Search in reactions')
        parser_reaction.set_defaults(which='reaction')
        parser_reaction.add_argument(
            '--id', '-i', dest='id', metavar='id',
            action=FilePrefixAppendAction, type=str, default=[],
            help='Reaction ID')
        parser_reaction.add_argument(
            '--compound', '-c', dest='compound', metavar='compound',
            action=FilePrefixAppendAction, type=str, default=[],
            help='Comma-separated list of compound IDs')

    def run(self):
        """Run search command."""

        which_command = self._args.which
        if which_command == 'compound':
            self._search_compound()
        elif which_command == 'reaction':
            self._search_reaction()

    def _search_compound(self):
        selected_compounds = set()

        for compound in self._model.parse_compounds():
            if len(self._args.id) > 0:
                if any(c == compound.id for c in self._args.id):
                    selected_compounds.add(compound)
                    continue

            if len(self._args.name) > 0:
                names = set()
                if 'name' in compound.properties:
                    names.add(compound.properties['name'])
                names.update(compound.properties.get('names', []))
                names = set(filter_search_term(n) for n in names)
                if any(filter_search_term(n) in names
                       for n in self._args.name):
                    selected_compounds.add(compound)
                    continue

        # Show results
        for compound in selected_compounds:
            props = set(compound.properties) - {'id'}
            print('id: {}'.format(compound.id))
            for prop in sorted(props):
                print('{}: {}'.format(prop, compound.properties[prop]))
            if compound.filemark is not None:
                print('Defined in {}'.format(compound.filemark))
            print()

    def _search_reaction(self):
        selected_reactions = set()

        # Prepare translation table from compound id to name
        compound_name = {}
        for compound in self._model.parse_compounds():
            if 'name' in compound.properties:
                compound_name[compound.id] = compound.properties['name']

        # Prepare sets of matched compounds
        search_compounds = []
        for compound_list in self._args.compound:
            compound_set = set()
            for compound_spec in compound_list.split(','):
                compound_set.add(parse_compound(compound_spec.strip()))
            search_compounds.append(compound_set)

        for reaction in self._model.parse_reactions():
            if len(self._args.id) > 0:
                if any(r == reaction.id for r in self._args.id):
                    selected_reactions.add(reaction)
                    continue

            if len(search_compounds) > 0:
                compounds = set(c for c, _ in reaction.equation.compounds)
                compounds.update(c.in_compartment(None) for c, _ in
                                 reaction.equation.compounds)
                if any(c.issubset(compounds) for c in search_compounds):
                    selected_reactions.add(reaction)
                    continue

        # Show results
        for reaction in selected_reactions:
            props = set(reaction.properties) - {'id', 'equation'}
            print('id: {}'.format(reaction.id))
            print('equation: {}'.format(
                reaction.equation))
            translated_equation = reaction.equation.translated_compounds(
                lambda x: compound_name.get(x, x))
            if reaction.equation != translated_equation:
                print('equation (compound names): {}'.format(
                    translated_equation))
            for prop in sorted(props):
                print('{}: {}'.format(prop, reaction.properties[prop]))
            if reaction.filemark is not None:
                print('Defined in {}'.format(reaction.filemark))
            print()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import time
import logging

from six import iteritems

from ..command import (Command, CommandError, SolverCommandMixin,
                       FilePrefixAppendAction)
from .. import massconsistency
from ..reaction import Compound

logger = logging.getLogger(__name__)


class MassConsistencyCommand(SolverCommandMixin, Command):
    """Check whether the model is mass consistent."""

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--exclude', metavar='reaction', action=FilePrefixAppendAction,
            type=str, default=[],
            help='Exclude reaction from mass consistency')
        parser.add_argument(
            '--epsilon', type=float, help='Mass threshold',
            default=1e-5)
        parser.add_argument(
            '--checked', metavar='reaction', action=FilePrefixAppendAction,
            type=str, default=[],
            help='Mark reaction as already checked (no residual)')
        parser.add_argument(
            '--type', choices=['compound', 'reaction'],
            default='compound', help='Type of check to perform')
        super(MassConsistencyCommand, cls).init_parser(parser)

    def run(self):
        # Load compound information
        self._compound_name = {}
        zeromass = set()
        for compound in self._model.parse_compounds():
            if 'name' in compound.properties:
                self._compound_name[compound.id] = compound.properties['name']
            elif compound.id not in self._compound_name:
                self._compound_name[compound.id] = compound.id

            if compound.properties.get('zeromass', False):
                zeromass.add(Compound(compound.id))

        # Create a set of known mass-inconsistent reactions
        exchange = set()
        for reaction_id in self._mm.reactions:
            if self._mm.is_exchange(reaction_id):
                exchange.add(reaction_id)

        # Create a set of excluded reactions
        exclude = set(self._args.exclude)

        # Add biomass reaction to be excluded
        biomass_reaction = self._model.get_biomass_reaction()
        if biomass_reaction is not None:
            exclude.add(biomass_reaction)

        solver = self._get_solver()

        known_inconsistent = exclude | exchange

        if self._args.type == 'compound':
            self._check_compounds(known_inconsistent, zeromass, solver)
        elif self._args.type == 'reaction':
            self._check_reactions(known_inconsistent, zeromass, solver)
        else:
            raise CommandError('Invalid type of check: {}'.format(
                self._args.type))

    def _check_compounds(self, known_inconsistent, zeromass, solver):
        logger.info('Checking stoichiometric consistency of compounds...')

        epsilon = self._args.epsilon

        start_time = time.time()

        masses = dict(massconsistency.check_compound_consistency(
            self._mm, solver, known_inconsistent, zeromass))

        logger.info('Solving took {:.2f} seconds'.format(
            time.time() - start_time))

        good = 0
        total = 0
        for compound, mass in sorted(
                iteritems(masses), key=lambda x: (x[1], x[0]), reverse=True):
            if mass >= 1-epsilon or compound.name in zeromass:
                good += 1
            total += 1
            print('{}\t{}\t{}'.format(
                compound, mass, compound.translate(
                    lambda x: self._compound_name.get(x, x))))
        logger.info('Consistent compounds: {}/{}'.format(good, total))

    def _check_reactions(self, known_inconsistent, zeromass, solver):
        logger.info('Checking stoichiometric consistency of reactions...')

        # Create a set of checked reactions
        checked = set(self._args.checked)

        epsilon = self._args.epsilon

        start_time = time.time()

        reaction_iter, compound_iter = (
            massconsistency.check_reaction_consistency(
                self._mm, solver=solver, exchange=known_inconsistent,
                checked=checked, zeromass=zeromass))

        logger.info('Solving took {:.2f} seconds'.format(
            time.time() - start_time))

        good = 0
        total = 0
        for reaction_id, residual in sorted(
                reaction_iter, key=lambda x: abs(x[1]), reverse=True):
            total += 1
            if abs(residual) >= epsilon:
                reaction = self._mm.get_reaction(reaction_id)
                rxt = reaction.translated_compounds(
                    lambda x: self._compound_name.get(x, x))
                print('{}\t{}\t{}'.format(reaction_id, residual, rxt))
            else:
                good += 1
        logger.info('Consistent reactions: {}/{}'.format(good, total))



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

from setuptools import setup, find_packages

# Read long description
with open('README.rst') as f:
    long_description = f.read()

setup(
    name='psamm',
    version='0.16',
    description='PSAMM metabolic modeling tools',
    maintainer='Jon Lund Steffensen',
    maintainer_email='jon_steffensen@uri.edu',
    url='https://github.com/zhanglab/psamm',
    license='GNU GPLv3+',

    long_description=long_description,

    classifiers = [
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4'
    ],

    packages=find_packages(),
    entry_points = {
        'console_scripts': [
            'psamm-model = psamm.command:main',
            'psamm-list-lpsolvers = psamm.lpsolver.generic:list_solvers'
        ],
        'psamm.commands': [
            'chargecheck = psamm.commands.chargecheck:ChargeBalanceCommand',
            'console = psamm.commands.console:ConsoleCommand',
            'excelexport = psamm.commands.excelexport:ExcelExportCommand',
            'fastgapfill = psamm.commands.fastgapfill:FastGapFillCommand',
            'fba = psamm.commands.fba:FluxBalanceCommand',
            'fluxcheck = psamm.commands.fluxcheck:FluxConsistencyCommand',
            'fluxcoupling = psamm.commands.fluxcoupling:FluxCouplingCommand',
            'formulacheck = psamm.commands.formulacheck:FormulaBalanceCommand',
            'fva = psamm.commands.fva:FluxVariabilityCommand',
            'gapfill = psamm.commands.gapfill:GapFillCommand',
            'masscheck = psamm.commands.masscheck:MassConsistencyCommand',
            'randomsparse = psamm.commands.randomsparse:RandomSparseNetworkCommand',
            'robustness = psamm.commands.robustness:RobustnessCommand',
            'sbmlexport = psamm.commands.sbmlexport:SBMLExport',
            'search = psamm.commands.search:SearchCommand',
            'tableexport = psamm.commands.tableexport:ExportTableCommand',
        ]
    },

    test_suite='psamm.tests',

    install_requires=[
        'PyYAML>=3.11,<4.0',
        'six'
    ],
    extras_require={
        'docs': ['sphinx', 'mock']
    })



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

"""Representation of metabolic network models."""

from collections import Mapping

from .database import MetabolicDatabase, StoichiometricMatrixView
from .reaction import Reaction


class FluxBounds(object):
    """Represents lower and upper bounds of flux as a mutable object

    This object is used internally in the model representation. Changing
    the state of the object will change the underlying model parameters.
    Deleting a value will reset that value to the defaults.
    """

    def __init__(self, model, reaction):
        self._model = model
        self._reaction = reaction

    def __iter__(self):
        '''Iterator over lower and upper value'''
        yield self.lower
        yield self.upper

    def _assign_lower(self, value):
        self._model._limits_lower[self._reaction] = value

    def _assign_upper(self, value):
        self._model._limits_upper[self._reaction] = value

    def _assign_both(self, lower, upper):
        self._assign_lower(lower)
        self._assign_upper(upper)

    def _check_bounds(self, lower, upper):
        if lower > upper:
            raise ValueError('Lower bound larger than upper bound')

    @property
    def lower(self):
        '''Lower bound'''
        try:
            return self._model._limits_lower[self._reaction]
        except KeyError:
            if self._model.is_reversible(self._reaction):
                return -self._model._v_max
            else:
                return 0

    @lower.setter
    def lower(self, value):
        self._check_bounds(value, self.upper)
        self._assign_lower(value)

    @lower.deleter
    def lower(self):
        self._model._limits_lower.pop(self._reaction, None)

    @property
    def upper(self):
        '''Upper bound'''
        try:
            return self._model._limits_upper[self._reaction]
        except KeyError:
            return self._model._v_max

    @upper.setter
    def upper(self, value):
        self._check_bounds(self.lower, value)
        self._assign_upper(value)

    @upper.deleter
    def upper(self):
        self._model._limits_upper.pop(self._reaction, None)

    @property
    def bounds(self):
        '''Bounds as a tuple'''
        return self.lower, self.upper

    @bounds.setter
    def bounds(self, value):
        lower, upper = value
        self._check_bounds(lower, upper)
        self._assign_both(lower, upper)

    @bounds.deleter
    def bounds(self):
        del self.lower
        del self.upper

    def __eq__(self, other):
        """Equality test"""
        return (isinstance(other, FluxBounds) and
                self.lower == other.lower and
                self.upper == other.upper)

    def __ne__(self, other):
        """Inequality test"""
        return not self == other

    def __repr__(self):
        return '{}({}, {})'.format(
            self.__class__.__name__, repr(self.lower), repr(self.upper))


class LimitsView(Mapping):
    """Provides a view of the flux bounds defined in the model

    This object is used internally in MetabolicModel to
    expose a dictonary view of the FluxBounds associated
    with the model reactions.
    """

    def __init__(self, model):
        super(LimitsView, self).__init__()
        self._model = model

    def _create_bounds(self, reaction):
        return FluxBounds(self._model, reaction)

    def __getitem__(self, key):
        if not self._model.has_reaction(key):
            raise KeyError(key)
        return self._create_bounds(key)

    def __iter__(self):
        return self._model.reactions

    def __len__(self):
        return sum(1 for _ in self._model.reactions)


class MetabolicModel(MetabolicDatabase):
    """Represents a metabolic model containing a set of reactions

    The model contains a list of reactions referencing the reactions
    in the associated database.
    """

    def __init__(self, database, v_max=1000):
        self._database = database
        self._limits_lower = {}
        self._limits_upper = {}

        self._reaction_set = set()
        self._compound_set = set()

        self._v_max = v_max

    @property
    def database(self):
        return self._database

    @property
    def reactions(self):
        return iter(self._reaction_set)

    @property
    def compounds(self):
        return iter(self._compound_set)

    @property
    def compartments(self):
        compartment_set = set()
        for compound in self.compounds:
            if compound.compartment not in compartment_set:
                compartment_set.add(compound.compartment)
                yield compound.compartment

    def has_reaction(self, reaction_id):
        return reaction_id in self._reaction_set

    def get_reaction(self, reaction_id):
        if reaction_id not in self._reaction_set:
            raise ValueError('Reaction not in model: {}'.format(reaction_id))
        return self._database.get_reaction(reaction_id)

    def get_reaction_values(self, reaction_id):
        """Return stoichiometric values of reaction as a dictionary"""
        if reaction_id not in self._reaction_set:
            raise ValueError('Unknown reaction: {}'.format(repr(reaction_id)))
        return self._database.get_reaction_values(reaction_id)

    def get_compound_reactions(self, compound_id):
        """Iterate over all reaction ids the includes the given compound"""
        if compound_id not in self._compound_set:
            raise ValueError('Compound not in model: {}'.format(compound_id))

        for reaction_id in self._database.get_compound_reactions(compound_id):
            if reaction_id in self._reaction_set:
                yield reaction_id

    def is_reversible(self, reaction_id):
        """Whether the given reaction is reversible"""
        if reaction_id not in self._reaction_set:
            raise ValueError('Reaction not in model: {}'.format(reaction_id))
        return self._database.is_reversible(reaction_id)

    def is_exchange(self, reaction_id):
        """Whether the given reaction is an exchange reaction"""
        reaction = self.get_reaction(reaction_id)
        return len(reaction.left) == 0 or len(reaction.right) == 0

    @property
    def limits(self):
        return LimitsView(self)

    def add_reaction(self, reaction_id):
        """Add reaction to model"""

        if reaction_id in self._reaction_set:
            return

        reaction = self._database.get_reaction(reaction_id)
        self._reaction_set.add(reaction_id)
        for compound, _ in reaction.compounds:
            self._compound_set.add(compound)

    def remove_reaction(self, reaction):
        """Remove reaction from model"""

        if reaction not in self._reaction_set:
            return

        self._reaction_set.remove(reaction)
        self._limits_lower.pop(reaction, None)
        self._limits_upper.pop(reaction, None)

        # Remove compound from compound_set if it is not referenced
        # by any other reactions in the model.
        for compound, value in self._database.get_reaction_values(reaction):
            reactions = frozenset(
                self._database.get_compound_reactions(compound))
            if all(other_reaction not in self._reaction_set
                   for other_reaction in reactions):
                self._compound_set.remove(compound)

    def add_all_database_reactions(self, compartments={None, 'e'}):
        """Add all reactions from database that occur in given compartments"""

        added = set()
        for rxnid in self._database.reactions:
            reaction = self._database.get_reaction(rxnid)
            if all(compound.compartment in compartments
                   for compound, _ in reaction.compounds):
                if rxnid not in self._reaction_set:
                    added.add(rxnid)
                self.add_reaction(rxnid)

        return added

    def add_all_exchange_reactions(self, allow_duplicates=False):
        """Add all exchange reactions to database and to model"""

        all_reactions = {}
        if not allow_duplicates:
            # TODO: Avoid adding reactions that already exist in the database.
            # This should be integrated in the database.
            for rxnid in self._database.reactions:
                rx = self._database.get_reaction(rxnid)
                all_reactions[rx] = rxnid

        added = set()
        for compound in sorted(self.compounds):
            rxnid_ex = ('rxnex', compound)
            if not self._database.has_reaction(rxnid_ex):
                reaction_ex = Reaction(
                    Reaction.Bidir, [(compound.in_compartment('e'), 1)], [])
                if reaction_ex not in all_reactions:
                    self._database.set_reaction(rxnid_ex, reaction_ex)
                else:
                    rxnid_ex = all_reactions[reaction_ex]

            if rxnid_ex not in self._reaction_set:
                added.add(rxnid_ex)
            self.add_reaction(rxnid_ex)

        return added

    def add_all_transport_reactions(self, allow_duplicates=False):
        """Add all transport reactions to database and to model"""

        all_reactions = {}
        if not allow_duplicates:
            # TODO: Avoid adding reactions that already exist in the database.
            # This should be integrated in the database.
            for rxnid in self._database.reactions:
                rx = self._database.get_reaction(rxnid)
                all_reactions[rx] = rxnid

        added = set()
        for compound in sorted(self.compounds):
            if compound.compartment == 'e':
                # A transport reaction with exchange would not be valid
                continue

            rxnid_tp = ('rxntp', compound)
            if not self._database.has_reaction(rxnid_tp):
                reaction_tp = Reaction(
                    Reaction.Bidir,
                    [(compound.in_compartment('e'), 1)],
                    [(compound, 1)])
                if reaction_tp not in all_reactions:
                    self._database.set_reaction(rxnid_tp, reaction_tp)
                else:
                    rxnid_tp = all_reactions[reaction_tp]

            if rxnid_tp not in self._reaction_set:
                added.add(rxnid_tp)
            self.add_reaction(rxnid_tp)

        return added

    def copy(self):
        """Return copy of model"""

        model = self.__class__(self._database)
        model._limits_lower = dict(self._limits_lower)
        model._limits_upper = dict(self._limits_upper)
        model._reaction_set = set(self._reaction_set)
        model._compound_set = set(self._compound_set)
        return model

    @classmethod
    def load_model(cls, database, reaction_iter, medium=None, limits=None,
                   v_max=None):
        """Get model from reaction name iterator

        The model will contain all reactions of the iterator.
        """

        model_args = {}
        if v_max is not None:
            model_args['v_max'] = v_max

        model = cls(database, **model_args)
        for reaction_id in reaction_iter:
            model.add_reaction(reaction_id)

        # Apply reaction limits
        if limits is not None:
            for reaction_id, lower, upper in limits:
                if lower is not None:
                    model.limits[reaction_id].lower = lower
                if upper is not None:
                    model.limits[reaction_id].upper = upper

        # TODO: Currently we just create a new exchange reaction in the
        # database and add it to the model. Ideally, we should not modify
        # the database. The exchange reaction could be created on the
        # fly when required.
        if medium is not None:
            for compound, reaction_id, lower, upper in medium:
                # Create exchange reaction
                if reaction_id is None:
                    reaction_id = 'EX_{}_{}'.format(
                        compound.name, compound.compartment)
                model.database.set_reaction(
                    reaction_id, Reaction(Reaction.Bidir, [(compound, 1)], []))
                model.add_reaction(reaction_id)
                if lower is not None:
                    model.limits[reaction_id].lower = lower
                if upper is not None:
                    model.limits[reaction_id].upper = upper

        return model


class FlipableFluxBounds(FluxBounds):
    """FluxBounds object for a FlipableModelView.

    This object is used internally in the FlipableModelView to represent
    the bounds of flux on a reaction that can be flipped.
    """

    def __init__(self, view, reaction):
        super(FlipableFluxBounds, self).__init__(view._model, reaction)
        self._view = view

    def _assign_lower(self, value):
        if self._reaction in self._view._flipped:
            super(FlipableFluxBounds, self)._assign_upper(-value)
        else:
            super(FlipableFluxBounds, self)._assign_lower(value)

    def _assign_upper(self, value):
        if self._reaction in self._view._flipped:
            super(FlipableFluxBounds, self)._assign_lower(-value)
        else:
            super(FlipableFluxBounds, self)._assign_upper(value)

    @property
    def lower(self):
        """Lower bound"""
        if self._reaction in self._view._flipped:
            return -super(FlipableFluxBounds, self).upper
        return super(FlipableFluxBounds, self).lower

    @property
    def upper(self):
        """Upper bound"""
        if self._reaction in self._view._flipped:
            return -super(FlipableFluxBounds, self).lower
        return super(FlipableFluxBounds, self).upper


class FlipableStoichiometricMatrixView(StoichiometricMatrixView):
    """Provides a matrix view that flips with the flipable model view.

    This object is used internally in FlipableModelView to
    expose a matrix view that negates the stoichiometric
    values of flipped reactions.
    """

    def __init__(self, view):
        super(FlipableStoichiometricMatrixView, self).__init__(view._model)
        self._view = view

    def _value_mul(self, reaction):
        return -1 if reaction in self._view._flipped else 1

    def __getitem__(self, key):
        if len(key) != 2:
            raise KeyError(key)
        compound, reaction = key
        orig_value = (
            super(FlipableStoichiometricMatrixView, self).__getitem__(key))
        return self._value_mul(reaction) * orig_value


class FlipableLimitsView(LimitsView):
    """Provides a limits view that flips with the flipable model view.

    This object is used internally in FlipableModelView to
    expose a limits view that flips the bounds of all flipped
    reactions.
    """

    def __init__(self, view):
        super(FlipableLimitsView, self).__init__(view._model)
        self._view = view

    def _create_bounds(self, reaction):
        return FlipableFluxBounds(self._view, reaction)


class FlipableModelView(object):
    """Proxy wrapper of model objects allowing a flipped set of reactions.

    The proxy will forward all properties normally except
    that flipped reactions will appear to have stoichiometric
    values negated in the matrix property, and have bounds in
    the limits property flipped. This view is needed for
    some algorithms.
    """

    def __init__(self, model, flipped=set()):
        self._model = model
        self._flipped = set(flipped)

    @property
    def matrix(self):
        return FlipableStoichiometricMatrixView(self)

    @property
    def limits(self):
        return FlipableLimitsView(self)

    def flip(self, subset):
        self._flipped ^= subset

    def __getattr__(self, name):
        return getattr(self._model, name)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.expression.boolean import (Expression, And, Or, Variable,
                                      ParseError, SubstitutionError)


class TestVariable(unittest.TestCase):
    def test_variable_init(self):
        v = Variable('x')

    def test_variable_init_long_symbol(self):
        v = Variable('xyz')

    def test_variable_init_number_symbol(self):
        v = Variable('x2')

    def test_variable_init_underscore_symbol(self):
        v = Variable('x_y')

    def test_variable_init_from_unicode(self):
        symbol = u'\u00c6\u00d8\u00c5'
        v = Variable(symbol)
        self.assertEqual(v.symbol, symbol)

    def test_variable_init_with_dot_symbol(self):
        v = Variable('x12345.6')

    def test_variable_init_with_number(self):
        v = Variable('123')

    def test_variable_symbol(self):
        self.assertEqual(Variable('x').symbol, 'x')

    def test_variables_equals_true(self):
        self.assertEqual(Variable('x'), Variable('x'))

    def test_variables_not_equals(self):
        self.assertNotEqual(Variable('x'), Variable('y'))

    def test_variable_not_equals_bool(self):
        self.assertNotEqual(Variable('x'), True)

    def test_variable_hash(self):
        self.assertEqual(hash(Variable('xyz')), hash(Variable('xyz')))


class TestExpression(unittest.TestCase):
    def test_expression_substitute_existing(self):
        e = Expression('b1')
        e1 = e.substitute(lambda v: {'b1': True}.get(v.symbol, v))
        self.assertTrue(e1.has_value())
        self.assertTrue(e1.value)

    def test_expression_substitute_existing_false(self):
        e = Expression('b1')
        e1 = e.substitute(lambda v: {'b1': False}.get(v.symbol, v))
        self.assertTrue(e1.has_value())
        self.assertFalse(e1.value)

    def test_expression_substitute_unknown_to_variable(self):
        e = Expression('b1')
        e1 = e.substitute(lambda v: v)
        self.assertEqual(e1, Expression(Variable('b1')))

    def test_expression_substitute_unknown_to_expression(self):
        e = Expression('b1 and b2')
        e1 = e.substitute(lambda v: v)
        self.assertEqual(e1, Expression('b1 and b2'))

    def test_expression_substitute_invalid(self):
        e = Expression('b1 and b2')
        with self.assertRaises(SubstitutionError):
            e.substitute(lambda v: {'b1': 17}.get(v.symbol, v))

    def test_expression_substitute_with_short_circuit_and(self):
        e = Expression('a and (b or c) and d')
        e1 = e.substitute(lambda v: {'a': False}.get(v.symbol, v))
        self.assertTrue(e1.has_value())
        self.assertFalse(e1.value)

    def test_expression_substitute_with_short_circuit_or(self):
        e = Expression('(a and b) or c or d')
        e1 = e.substitute(lambda v: {'c': True}.get(v.symbol, v))
        self.assertTrue(e1.has_value())
        self.assertTrue(e1.value)

    def test_expression_substitute_remove_terms_from_and(self):
        e = Expression('a and b')
        e1 = e.substitute(lambda v: {'a': True}.get(v.symbol, v))
        self.assertEqual(e1, Expression(Variable('b')))

    def test_expression_substitute_remove_terms_from_or(self):
        e = Expression('a or b')
        e1 = e.substitute(lambda v: {'a': False}.get(v.symbol, v))
        self.assertEqual(e1, Expression(Variable('b')))

    def test_expression_substitute_evaluate_all_and_terms_to_true(self):
        e = Expression('a and b')
        e1 = e.substitute(lambda v: True)
        self.assertEqual(e1, Expression(True))

    def test_expression_substitute_evaluate_all_or_terms_to_false(self):
        e = Expression('a or b')
        e1 = e.substitute(lambda v: False)
        self.assertEqual(e1, Expression(False))

    def test_expression_iter_variables(self):
        e = Expression(
            Or(
                And(Variable('b1'), Variable('b2')),
                And(Variable('b3'), Variable('b4'))))
        self.assertEqual(list(e.variables), [
            Variable('b1'), Variable('b2'), Variable('b3'), Variable('b4')
        ])

    def test_expression_to_string(self):
        e = Expression('(a and b) or (c and d)')
        self.assertEqual(str(e), '(a and b) or (c and d)')

    def test_expression_with_variable_to_string(self):
        e = Expression(Variable('a'))
        self.assertEqual(str(e), 'a')

    def test_expression_with_bool_to_string(self):
        e = Expression(False)
        self.assertEqual(str(e), 'False')

    def test_expression_parse_and(self):
        e = Expression('b1 and b2')
        self.assertEqual(e, Expression(
            And(Variable('b1'), Variable('b2'))))

    def test_expression_parse_or(self):
        e = Expression('b1 or b2')
        self.assertEqual(e, Expression(
            Or(Variable('b1'), Variable('b2'))))

    def test_expression_parse_multiple_with_duplicates(self):
        e = Expression('b1 and b2 and b1 and b4')
        self.assertEqual(e, Expression(
            And(Variable('b1'), Variable('b2'), Variable('b4'))))

    def test_expression_parse_multiple_and(self):
        e = Expression('b1 and b2 and b3 and b4')
        self.assertEqual(e, Expression(
            And(
                Variable('b1'), Variable('b2'), Variable('b3'),
                Variable('b4'))))

    def test_expression_parse_multiple_or(self):
        e = Expression('b1 or b2 or b3 or b4')
        self.assertEqual(e, Expression(
            Or(
                Variable('b1'), Variable('b2'), Variable('b3'),
                Variable('b4'))))

    def test_expression_parse_multiple_parenthesis_and(self):
        e = Expression('b1 and (b2 and b3) and b4')
        self.assertEqual(e, Expression(
            And(
                Variable('b1'), Variable('b2'), Variable('b3'),
                Variable('b4'))))

    def test_expression_parse_multiple_parenthesis_or(self):
        e = Expression('b1 or (b2 or b3) or b4')
        self.assertEqual(e, Expression(
            Or(
                Variable('b1'), Variable('b2'), Variable('b3'),
                Variable('b4'))))

    def test_expression_parse_parentheses_mixed_1(self):
        e = Expression('(b1 and b2) or (b3 and b4)')
        self.assertEqual(e, Expression(
            Or(
                And(Variable('b1'), Variable('b2')),
                And(Variable('b3'), Variable('b4')))))

    def test_expression_parse_parentheses_mixed_2(self):
        e = Expression('(b1 or b2) and (b3 or b4)')
        self.assertEqual(e, Expression(
            And(
                Or(Variable('b1'), Variable('b2')),
                Or(Variable('b3'), Variable('b4')))))

    def test_expression_parse_uppercase_operators(self):
        e = Expression('(b1 AND b2) OR b3')
        self.assertEqual(e, Expression(
            Or(
                Variable('b3'),
                And(Variable('b1'), Variable('b2')))))

    def test_expression_parse_parenthesis_with_space_right(self):
        e = Expression('(b1 and b2 )')
        self.assertEqual(e, Expression(
            And(Variable('b1'), Variable('b2'))))

    def test_expression_parse_parenthesis_with_space_left(self):
        e = Expression('( b1 and b2)')
        self.assertEqual(e, Expression(
            And(Variable('b1'), Variable('b2'))))

    def test_expression_parse_parentheses_right_nested(self):
        e = Expression('(b1 or (b2 and (b3 or (b4 and (b5)))))')
        self.assertEqual(e, Expression(
            Or(
                Variable('b1'),
                And(
                    Variable('b2'),
                    Or(
                        Variable('b3'),
                        And(Variable('b4'), Variable('b5')))))))

    def test_expression_parse_parentheses_left_nested(self):
        e = Expression('(((b1 or b2) and b3) or b4) and (b5)')
        self.assertEqual(e, Expression(
            And(
                Variable('b5'),
                Or(
                    Variable('b4'),
                    And(
                        Variable('b3'),
                        Or(Variable('b1'), Variable('b2')))))))

    def test_expression_parse_implicit_mixed_1(self):
        e = Expression('b1 and b2 or b3 and b4')
        self.assertEqual(e, Expression(
            Or(
                And(Variable('b1'), Variable('b2')),
                And(Variable('b3'), Variable('b4')))))

    def test_expression_parse_implicit_mixed_2(self):
        e = Expression('b1 or b2 and b3 or b4')
        self.assertEqual(e, Expression(
            Or(
                Variable('b1'),
                And(Variable('b2'), Variable('b3')),
                Variable('b4'))))

    def test_expression_parse_longer_names(self):
        e = Expression('b12345 and bxyz and testtesttest')
        self.assertEqual(e, Expression(
            And(
                Variable('b12345'), Variable('bxyz'),
                Variable('testtesttest'))))

    def test_expression_parse_with_missing_variable(self):
        with self.assertRaises(ParseError):
            e = Expression('b1 and and b3')

    def test_expression_parse_with_missing_end_parenthesis(self):
        with self.assertRaises(ParseError):
            e = Expression('b1 and (b2 or b3')

    def test_expression_parse_name_starting_with_or(self):
        e = Expression('b1 and order')
        self.assertEqual(e, Expression(
            And(Variable('b1'), Variable('order'))))

    def test_expression_parse_name_starting_with_and(self):
        e = Expression('b1 or anderson')
        self.assertEqual(e, Expression(
            Or(Variable('b1'), Variable('anderson'))))

    def test_expression_parse_with_extra_space(self):
        e = Expression('b1    and   b2       and b3')
        self.assertEqual(e, Expression(
            And(Variable('b1'), Variable('b2'), Variable('b3'))))

    def test_expression_parse_with_square_mixed_groups(self):
        e = Expression('[(a or b) or (c or d)] or [e or (f and g and h)]')
        self.assertEqual(e, Expression(
            Or(
                Variable('a'), Variable('b'), Variable('c'), Variable('d'),
                Variable('e'),
                And(Variable('f'), Variable('g'), Variable('h')))))

    def test_expression_parse_with_square_groups_unmatched(self):
        with self.assertRaises(ParseError):
            e = Expression('[(a or b) or (c or d])')


class TestParseError(unittest.TestCase):
    def test_parse_error_without_indicator(self):
        e = ParseError('Random error')
        self.assertIsNone(e.indicator)

    def test_parse_error_with_indicator(self):
        e = ParseError('Random error', span=(3, 5))
        self.assertEqual(e.indicator, '   ^^')

    def test_parse_error_with_zero_width_indicator(self):
        e = ParseError('Random error', span=(4, 4))
        self.assertEqual(e.indicator, '    ^')


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import math
import logging

from ..command import Command, FilePrefixAppendAction

logger = logging.getLogger(__name__)


class ChargeBalanceCommand(Command):
    """Check whether compound charge is balanced.

    Balanced reactions are those reactions where the total charge
    is consistent on the left and right side of the reaction equation.
    Reactions that are not balanced will be printed out.
    """

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--exclude', metavar='reaction', action=FilePrefixAppendAction,
            type=str, default=[], help='Exclude reaction from balance check')
        parser.add_argument(
            '--epsilon', metavar='epsilon', type=float, default=1e-6,
            help='Threshold for charge imbalance to be considered zero'
        )
        super(ChargeBalanceCommand, cls).init_parser(parser)

    def run(self):
        """Run charge balance command"""

        # Load compound information
        compound_name = {}
        compound_charge = {}
        for compound in self._model.parse_compounds():
            compound_name[compound.id] = (
                compound.name if compound.name is not None else compound.id)
            if hasattr(compound, 'charge') and compound.charge is not None:
                compound_charge[compound.id] = compound.charge

        # Create a set of known charge-inconsistent reactions
        exchange = set()
        for reaction_id in self._mm.reactions:
            if self._mm.is_exchange(reaction_id):
                exchange.add(reaction_id)

        # Create a set of excluded reactions
        exclude = set(self._args.exclude)

        def reaction_charges(reaction_id):
            for compound, value in self._mm.get_reaction_values(reaction_id):
                charge = compound_charge.get(compound.name, float('nan'))
                yield charge * float(value)

        count = 0
        unbalanced = 0
        unchecked = 0
        for reaction in sorted(self._mm.reactions):
            if reaction in exchange:
                continue

            count += 1

            if reaction in exclude:
                continue

            charge = sum(reaction_charges(reaction))
            if math.isnan(charge):
                logger.debug('Not checking reaction {};'
                             ' missing charge'.format(reaction))
                unchecked += 1
            elif abs(charge) > self._args.epsilon:
                unbalanced += 1
                rx = self._mm.get_reaction(reaction)
                rxt = rx.translated_compounds(
                    lambda x: compound_name.get(x, x))
                print('{}\t{}\t{}'.format(reaction, charge, rxt))

        logger.info('Unbalanced reactions: {}/{}'.format(unbalanced, count))
        logger.info('Unchecked reactions due to missing charge: {}/{}'.format(
            unchecked, count))
        logger.info('Reactions excluded from check: {}/{}'.format(
            len(exclude), count))



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.lpsolver import lp

try:
    from psamm.lpsolver import cplex
except ImportError:
    cplex = None

requires_solver = unittest.skipIf(cplex is None, 'Cplex solver not available')


@requires_solver
class TestCplexProblem(unittest.TestCase):
    def setUp(self):
        self.solver = cplex.Solver()

    def test_expression_to_string(self):
        prob = self.solver.create_problem()
        prob.define('x', 'y', lower=0, upper=10)
        expr = -3 * prob.var('x') + prob.var('y')
        self.assertEqual(str(expr), '-3*x + y')

    def test_expression_of_tuple_to_string(self):
        prob = self.solver.create_problem()
        prob.define(('x', 1), ('x', 2), lower=0, upper=10)
        expr = -prob.var(('x', 1)) + 2 * prob.var(('x', 2))
        self.assertEqual(str(expr), "-('x', 1) + 2*('x', 2)")

    def test_objective_reset_on_set_linear_objective(self):
        prob = self.solver.create_problem()
        prob.define('x', 'y', lower=0, upper=10)
        prob.add_linear_constraints(prob.var('x') + prob.var('y') <= 12)
        prob.set_objective_sense(lp.ObjectiveSense.Maximize)

        # Solve first time, maximize x
        prob.set_linear_objective(2*prob.var('x'))
        result = prob.solve()
        self.assertAlmostEqual(result.get_value('x'), 10)

        # Solve second time, maximize y
        # If the objective is not properly cleared,
        # the second solve will still maximize x.
        prob.set_linear_objective(prob.var('y'))
        result = prob.solve()
        self.assertAlmostEqual(result.get_value('y'), 10)

    def test_result_to_bool_conversion_on_optimal(self):
        """Run a feasible LP problem and check that the result is True"""
        prob = self.solver.create_problem()
        prob.define('x', 'y', lower=0, upper=10)
        prob.add_linear_constraints(prob.var('x') + prob.var('y') <= 12)
        prob.set_objective_sense(lp.ObjectiveSense.Maximize)

        prob.set_linear_objective(2*prob.var('x'))
        result = prob.solve()
        self.assertTrue(result)

    def test_result_to_bool_conversion_on_infeasible(self):
        """Run an infeasible LP problem and check that the result is False"""
        prob = self.solver.create_problem()
        prob.define('x', 'y', 'z', lower=0, upper=10)
        prob.add_linear_constraints(2*prob.var('x') == -prob.var('y'),
                                    prob.var('x') + prob.var('z') >= 6,
                                    prob.var('z') <= 3)
        prob.set_objective_sense(lp.ObjectiveSense.Maximize)
        prob.set_linear_objective(2*prob.var('x'))
        result = prob.solve()
        self.assertFalse(result)

    def test_constraint_delete(self):
        prob = self.solver.create_problem()
        prob.define('x', 'y', lower=0, upper=10)
        prob.add_linear_constraints(prob.var('x') + prob.var('y') <= 12)
        c1, = prob.add_linear_constraints(prob.var('x') <= 5)

        prob.set_objective_sense(lp.ObjectiveSense.Maximize)
        prob.set_linear_objective(prob.var('x'))

        result = prob.solve()
        self.assertAlmostEqual(result.get_value('x'), 5)

        # Delete constraint
        c1.delete()
        result = prob.solve()
        self.assertAlmostEqual(result.get_value('x'), 10)


if __name__ == '__main__':
    unittest.main()
